# -*- coding: utf-8 -*-
"""
Contains functions not directly related to word parsing, but used by the it.
"""


def recursively_extract(node, exfun, maxdepth=2):
    """
    Transform a html ul/ol tree into a python list tree.

    Converts a html node containing ordered and unordered lists and list items
    into an object of lists with tree-like structure. Leaves are retrieved by
    applying `exfun` function to the html nodes not containing any ul/ol list.

    Args:
        node: BeautifulSoup HTML node to traverse
        exfun: function to apply to every string node found
        maxdepth: maximal depth of lists to go in the node

    Returns:
        A tree-like python object composed of lists.


    Examples:

    >>> node_content = \
    '''
    <ol>
        <li>Hase</li>
        <li>Nase<ol><li>Eins</li><li>Zwei</li></ol></li>
    </ol>'''
    >>> node = BeautifulSoup(node_content, "lxml")
    >>> recursively_extract(node, lambda x: x)
    [<li>Hase</li>, [<li>Eins</li>, <li>Zwei</li>]]
    >>> recursively_extract(node, lambda x: x.get_text())
    ['Hase', ['Eins', 'Zwei']]
    """
    if node.name in ["ol", "ul"]:
        lilist = node
    else:
        lilist = node.ol or node.ul
    if lilist and maxdepth:
        # apply 'recursively_extract' to every 'li' node found under this node
        return [
            recursively_extract(li, exfun, maxdepth=maxdepth - 1)
            for li in lilist.find_all("li", recursive=False)
        ]
    # if this node doesn't contain 'ol' or 'ul' node, return the transformed
    # leaf (using the 'exfun' function)
    return exfun(node)


def clear_text(text):
    """
    Remove soft hyphens anywhere, and heading and trailing spaces.
    """
    return text.replace("\xad", "").strip()



# -*- coding: utf-8 -*-
"""
Contains module version constant
"""
import pkg_resources

try:
    __version__ = pkg_resources.get_distribution("duden").version
except pkg_resources.DistributionNotFound:
    __version__ = "unknown"



# -*- coding: utf-8 -*-
"""
Network requests-related functions
"""

import gzip
import string
from pathlib import Path

import bs4
import requests
from xdg.BaseDirectory import xdg_cache_home

from .common import clear_text
from .inflection import Inflector
from .word import DudenWord

URL_FORM = "https://www.duden.de/rechtschreibung/{word}"
SEARCH_URL_FORM = "https://www.duden.de/suchen/dudenonline/{word}"
GRAMMAR_BASE = "https://www.duden.de/{urlpart}"
DEFAULT_TIMEOUT = 10


def sanitize_word(word):
    """
    Sanitize unicode word for use as filename

    Ascii letters and underscore are kept unchanged.
    Other characters are replaced with "-u{charccode}-" string.
    """
    allowed_chars = string.ascii_letters + "_"

    def sanitize_char(char):
        if char in allowed_chars:
            return char
        return "-u" + str(ord(char)) + "-"

    return "".join(sanitize_char(char) for char in word)


def cached_response(prefix=""):
    """
    Add `cache=True` keyword argument to a function to allow result caching based on single string
    argument.
    """

    def decorator_itself(func):
        def function_wrapper(cache_key, cache=True, **kwargs):
            cachedir = Path(xdg_cache_home) / "duden"
            filename = prefix + sanitize_word(cache_key) + ".gz"
            full_path = str(cachedir / filename)

            if cache:
                # try to read from cache
                cachedir.mkdir(parents=True, exist_ok=True)
                try:
                    with gzip.open(full_path, "rt", encoding="utf8") as file:
                        return file.read()
                except (FileNotFoundError, IOError, EOFError):
                    pass

            result = func(cache_key, **kwargs)

            if cache and result is not None:
                with gzip.open(full_path, "wt", encoding="utf8") as file:
                    file.write(result)

            return result

        return function_wrapper

    return decorator_itself


@cached_response(prefix="")
def request_word(word):
    """
    Request word page from duden
    """
    url = URL_FORM.format(word=word)
    try:
        response = requests.get(url, timeout=DEFAULT_TIMEOUT)
    except requests.exceptions.ConnectionError as exc:
        raise RuntimeError(
            _("Connection could not be established. Check your internet connection.")
        ) from exc

    if response.status_code == 404:
        return None
    response.raise_for_status()

    return response.text


def get(word, cache=True):
    """
    Load the word 'word' and return the DudenWord instance
    """
    html_content = request_word(
        word, cache=cache
    )  # pylint: disable=unexpected-keyword-arg
    if html_content is None:
        return None

    soup = bs4.BeautifulSoup(html_content, "html.parser")
    return DudenWord(soup)


def get_word_of_the_day():
    """
    Scrapes the word of the day and returns DudenWord instance of it.
    """
    html_content = requests.get("https://www.duden.de", timeout=DEFAULT_TIMEOUT).content
    soup = bs4.BeautifulSoup(html_content, "html.parser")
    link = soup.find("a", class_="scene__title-link").get("href")
    word = link.split("/")[-1]  # get word from "/rechtschreibung/word"
    return get(word)


def get_search_link_variants(link_text):
    """
    Lists possible interpretations of link text on search page.

    Used for determining whether a search page entry matches the search term.
    """
    return clear_text(link_text).split(", ")


@cached_response(prefix="search-")
def request_search(word):
    """
    Request search page from duden
    """
    url = SEARCH_URL_FORM.format(word=word)
    return requests.get(url, timeout=DEFAULT_TIMEOUT).text


def search(word, exact=True, return_words=True, cache=True):
    """
    Search for a word 'word' in duden
    """
    response_text = request_search(
        word, cache=cache
    )  # pylint: disable=unexpected-keyword-arg
    soup = bs4.BeautifulSoup(response_text, "html.parser")
    definitions = soup.find_all("h2", class_="vignette__title")

    if definitions is None:
        return []

    urlnames = []
    for definition in definitions:
        definition_title = definition.text
        if (not exact) or word in get_search_link_variants(definition_title):
            urlnames.append(definition.find("a")["href"].split("/")[-1])

    if not return_words:
        return urlnames
    return [get(urlname, cache=cache) for urlname in urlnames]


@cached_response(prefix="grammar-")
def request_grammar(urlpart):
    """
    Fetch inflection-related page and cache the result

    Args:
        urlpart (str): the url suffix in the one of three forms:
            '/deklination/substantive/{word}'
            '/deklination/adjektive/{word}'
            '/konjugation/{word}'

    Returns:
        str: HTML content of the page
    """
    url = GRAMMAR_BASE.format(urlpart=urlpart)
    return requests.get(url, timeout=DEFAULT_TIMEOUT).text


def grammar(urlpart):
    """
    Return word inflections when given url suffix for word's grammar page

    Args:
        urlpart (str): the url suffix in the one of three forms:
            '/deklination/substantive/{word}'
            '/deklination/adjektive/{word}'
            '/konjugation/{word}'

    Returns:
        Inflector: object providing word inflections
    """
    response_text = request_grammar(urlpart)
    soup = bs4.BeautifulSoup(response_text, "html.parser")
    return Inflector(soup)



"""
Module containing the base class for duden pages
"""


class DudenPage:
    """
    Represents a general duden page
    """

    # pylint: disable=too-few-public-methods
    def __init__(self, soup):
        self.soup = soup

    def division(self, title):
        """Find general page division by its title"""
        for division in self.soup.find_all("div", class_="division"):
            div_title = division.find("h2", class_="division__title")
            if div_title and div_title.text == title:
                return division
        raise KeyError(title)



"""
Small module with functions around "grammar" pages like
/deklination/* and
/konjugation/*

The main class is GrammarPage representing the above pages.

Other module-level functions are helper functions to assist the class.
"""
import copy

from .base import DudenPage


class GrammarPage(DudenPage):
    """
    Represent pages like

    /deklination/{substantive,adjektive}/{word}
    /konjugation/{word}

    and do a rough parsing of the accordion style HTML elements
    """

    @property
    def table_data(self):
        """Parse grammar data for grammar page"""
        return parse_grammar(self.division("Grammatik"))


def parse_grammar(division):
    """Parse div.Grammatik > div.con-dec__wrapper"""
    return {
        igrp.h3.text: parse_igroup(igrp)
        for igrp in division.div(class_="con-dec__wrapper")
    }


def parse_igroup(igrp):
    """Parse div.Grammatik > div.con-dec__wrapper > accordion-table"""
    return [parse_actable(actable) for actable in igrp.div(class_="accordion-table")]


def parse_actable(actable):
    """Parse div.Grammatik > div.con-dec__wrapper > accordion-table > ul"""
    return [parse_ul(tag_ul) for tag_ul in actable.find_all("ul")]


def parse_ul(tag_ul):
    """Parse div.Grammatik > div.con-dec__wrapper > accordion-table > ul > li"""
    return [parse_li(tag_li) for tag_li in tag_ul.find_all("li")]


def parse_li(tag_li):
    """Parse li tag contents"""
    tag_li = copy.copy(tag_li)
    try:
        tag_li.sup.extract()
    except AttributeError:
        pass
    return tag_li.text



"""
Module containing classes and functions related to word inflection.

The most important are the grammatical category enums like Case, Gender,
and the Inflector class.
"""
from enum import Enum

from .page.grammar import GrammarPage

# key names of Inflector raw data dict
NOUN = "Deklination"
ADJ_STRONG = "Starke Beugung (ohne Artikel)"
ADJ_WEAK = "Schwache Beugung (mit Artikel)"
ADJ_MIXED = "Gemischte Beugung (mit ein, kein, Possessivpronomen u. a.)"
ADJ_COMPARE = "Steigerungsformen"
VERB_INDICATIVE = "Indikativ"
VERB_SUBJUNCTIVE_I = "Konjunktiv I"
VERB_SUBJUNCTIVE_II = "Konjunktiv II"
VERB_IMPERATIVE = "Imperativ"
VERB_INFINITIVE_FORMS = "Infinite Formen"


class Number(Enum):
    """Grammatical number"""

    SINGULAR = "Singular"
    PLURAL = "Plural"


class Case(Enum):
    """Grammatical case"""

    NOMINATIVE = "Nominativ"
    GENITIVE = "Genitiv"
    DATIVE = "Dativ"
    ACCUSATIVE = "Akkusativ"


class Gender(Enum):
    """Grammatical gender"""

    MASCULINE = "Maskulinum"
    FEMININE = "Femininum"
    NEUTER = "Neutrum"


class Degree(Enum):
    """Grammatical degree (for adjective comparison)"""

    POSITIVE = "Positiv"
    COMPARATIVE = "Komparativ"
    SUPERLATIVE = "Superlativ"


# TODO: reduce to three categories
class Person(Enum):
    """Grammatical person"""

    FIRST_SINGULAR = "ich"
    SECOND_SINGULAR = "du"
    THIRD_SINGULAR = "er/sie/es"
    FIRST_PLURAL = "wir"
    SECOND_PLURAL = "ihr"
    THIRD_PLURAL = "sie"


class Mood(Enum):
    """Grammatical mood"""

    INDICATIVE = VERB_INDICATIVE
    SUBJUNCTIVE_I = VERB_SUBJUNCTIVE_I
    SUBJUNCTIVE_II = VERB_SUBJUNCTIVE_II


class Tense(Enum):
    """Grammatical tense"""

    PRESENT = "Präsens"
    PAST = "Präteritum"
    PERFECT = "Perfekt"
    PAST_PERFECT = "Plusquamperfekt"
    FUTURE = "Futur I"
    FUTURE_PERFECT = "Futur II"


# TODO: rely only on number
class ImperativePerson(Enum):
    """Grammatical person (only for imperative forms of verbs)"""

    PERSON_2_SINGULAR = "2. Person Singular [du]"
    PERSON_2_PLURAL = "2. Person Plural [ihr]"


class InfinitiveForm(Enum):
    """Infinitive form of verb"""

    INFINITIVE_WITH_ZU = "Infinitiv mit zu"
    PARTICIPLE_I = "Partizip I"
    PARTICIPLE_II = "Partizip II"


class KeyChainError(KeyError):
    """Variant of KeyError which stores previously accessed keys of a nested dict"""

    def __init__(self, key, previous_keys):
        super().__init__(key)
        self.key = key
        self.previous_keys = previous_keys


class Enumdict:
    """
    A dict wrapper with these properties

    * Enums behave as their raw values when used as keys (d["Singular"] == d[Number.SINGULAR])
    * Nested inner dicts are also Enumdicts and store the key path they were accessed with
    """

    def __init__(self, source, key_prefix=None):
        self.source = source
        self.key_prefix = key_prefix or []  # used for error messages

    def __getitem__(self, key):
        real_key = key.value if isinstance(key, Enum) else key
        try:
            value = self.source[real_key]
        except KeyError:
            raise KeyChainError(real_key, self.key_prefix) from None
        new_prefix = self.key_prefix + [real_key]
        return (
            Enumdict(value, key_prefix=new_prefix) if isinstance(value, dict) else value
        )

    def __repr__(self):
        return f"{self.__class__.__name__}({repr(self.source)})"


class Inflector:
    """
    Provides methods for word inflection

    Example:

        > r = requests.get("https://www.duden.de/deklination/substantive/Hase")
        > soup = bs4.BeautifulSoup(r.text)
        > inf = Inflector(soup)
        > inf.noun_decline(Number.SINGULAR, Case.NOMINATIVE)
        'der Hase'
    """

    def __init__(self, soup):
        """
        Construct Inflector object

        Args:
            soup (BeautifulSoup): parsed grammar page
        """
        self.page = GrammarPage(soup)
        self.data = {
            key: conditional_transform(key, value)
            for key, value in self.page.table_data.items()
        }
        self.enumraw = Enumdict(self.data)

    def __repr__(self):
        if not self.data:
            example = "Empty"
        else:
            example = self.data
            while isinstance(example, dict):
                example = list(example.values())[0]
            example = repr(example) + ", ..."
        return f"({self.__class__.__name__}: {example})"

    def inflect(self, *key_chain):
        """Fetch data from nested dict in self.enumraw by providing a list of keys to apply"""
        inner = self.enumraw
        try:
            for key in key_chain:
                inner = inner[key]
        except KeyChainError as err:
            keys_str = ".".join(repr(key) for key in (err.previous_keys + [err.key]))
            other_choices = ", ".join(repr(key) for key in inner.source.keys())
            err_msg = _("Cannot inflect. Missing data for: {} . Did you mean {}?")
            raise ValueError(err_msg.format(keys_str, other_choices)) from None
        return inner

    # nouns
    def noun_decline(self, number, case):
        """
        number: "Singular" or "Plural"
        case: "Nominativ", "Genitiv", "Dativ", "Akkusativ"
        """
        return self.inflect(NOUN, number, case)

    # adjectives
    def adjective_decline_strong(self, gender, case):
        """Strong adjective inflection"""
        return self.inflect(ADJ_STRONG, gender, case)

    def adjective_decline_weak(self, gender, case):
        """Weak adjective inflection"""
        return self.inflect(ADJ_WEAK, gender, case)

    def adjective_decline_mixed(self, gender, case):
        """Mixed adjective inflection"""
        return self.inflect(ADJ_MIXED, gender, case)

    def adjective_compare(self, degree):
        """Adjective comparative"""
        return self.inflect(ADJ_COMPARE, degree)

    # verbs
    def verb_conjugate(self, mood, tense, person):
        """Verb conjugation"""
        return self.inflect(mood, tense, person)

    def verb_imperative(self, person):
        """Verb imperative form"""
        return self.inflect(VERB_IMPERATIVE, person)

    def verb_infinitive_forms(self, form):
        """Verb infinitive form"""
        return self.inflect(VERB_INFINITIVE_FORMS, form)


def legend_left_transform(structure):
    """Transform data for tables where legend on the left"""
    return dict(zip(*structure[0]))


def legend_top_transform(structure):
    """Transform data for tables where legend at the top"""
    return dict(s[0] for s in structure)


def hidden_title_transform(structure):
    """Transform data for most verb and noun tables"""
    resmap = {}
    for legend, content in structure:
        name = content[0]
        mapping = dict(zip(legend[1:], content[1:]))
        resmap[name] = mapping
    return resmap


def square_transform(structure):
    """Transform data for most adjective tables"""
    _, *left_legend = structure[0][0]
    res = {}
    for _, column in structure:
        top_key, *values = column
        res[top_key] = dict(zip(left_legend, values))
    return res


def conditional_transform(key, structure):
    """Transform grammar table data conditionally based on table title"""
    transformation = table_transformations[key]
    return transformation(structure)


# how individual top-level section table_data must be transformed
table_transformations = {
    # noun
    NOUN: hidden_title_transform,
    # verb
    VERB_INDICATIVE: hidden_title_transform,
    VERB_SUBJUNCTIVE_I: hidden_title_transform,
    VERB_SUBJUNCTIVE_II: hidden_title_transform,
    VERB_IMPERATIVE: legend_left_transform,
    VERB_INFINITIVE_FORMS: legend_top_transform,
    # adjective sections
    ADJ_COMPARE: legend_top_transform,
    ADJ_STRONG: square_transform,
    ADJ_WEAK: square_transform,
    ADJ_MIXED: square_transform,
}



# -*- coding: utf-8 -*-
"""
CLI related functions
"""

import argparse
import sys

import yaml
from crayons import blue, red, white  # pylint: disable=no-name-in-module

from .__version__ import __version__
from .display import (
    describe_word,
    display_compounds,
    display_inflections,
    print_string_or_list,
    print_tree_of_strings,
)
from .request import get, search


def display_word(word, args):
    """
    Display word attribute or general description, based on commandline arguments
    """
    # pylint: disable=too-many-branches, too-many-statements
    if args.title:
        print(word.title)
    elif args.name:
        print(word.name)
    elif args.article:
        if word.article:
            print(word.article)
    elif args.part_of_speech:
        if word.part_of_speech:
            print(word.part_of_speech)
    elif args.frequency:
        if word.frequency:
            print(word.frequency)
    elif args.usage:
        if word.usage:
            print(word.usage)
    elif args.word_separation:
        for part in word.word_separation:
            print(part)
    elif args.meaning_overview:
        if word.meaning_overview:
            print_tree_of_strings(word.meaning_overview)
    elif args.synonyms:
        synonyms = word.synonyms
        if synonyms:
            print_string_or_list(synonyms)
    elif args.origin:
        if word.origin:
            print(word.origin)
    elif args.grammar_overview:
        if word.grammar_overview:
            print(word.grammar_overview)
    elif args.compounds:
        if word.compounds:
            display_compounds(word, args.compounds)
    elif args.phonetic:
        if word.phonetic:
            print(word.phonetic)
    elif args.alternative_spellings:
        if word.alternative_spellings:
            for spelling in word.alternative_spellings:
                print(spelling)
    elif args.inflect:
        if word.inflection:
            display_inflections(word)
    elif args.export:
        yaml_string = yaml.dump(word.export(), sort_keys=False, allow_unicode=True)
        print(yaml_string, end="")
    elif args.words_before:
        print_string_or_list(word.words_before)
    elif args.words_after:
        print_string_or_list(word.words_after)
    else:
        # print the description
        describe_word(word)


def parse_args():
    """
    Parse CLI arguments
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("word")
    parser.add_argument(
        "--title", action="store_true", help=_("display word and article")
    )
    parser.add_argument(
        "--name", action="store_true", help=_("display the word itself")
    )
    parser.add_argument("--article", action="store_true", help=_("display article"))
    parser.add_argument(
        "--part-of-speech", action="store_true", help=_("display part of speech")
    )
    parser.add_argument(
        "--frequency", action="store_true", help=_("display commonness (1 to 5)")
    )
    parser.add_argument(
        "--usage", action="store_true", help=_("display context of use")
    )
    parser.add_argument(
        "--word-separation",
        action="store_true",
        help=_("display proper separation (line separated)"),
    )
    parser.add_argument(
        "--meaning-overview", action="store_true", help=_("display meaning overview")
    )
    parser.add_argument(
        "--synonyms", action="store_true", help=_("list synonyms (line separated)")
    )
    parser.add_argument("--origin", action="store_true", help=_("display origin"))
    parser.add_argument(
        "--grammar-overview",
        action="store_true",
        help=_("display short grammar overview"),
    )
    parser.add_argument(
        "--compounds", nargs="?", const="ALL", help=_("list common compounds")
    )
    parser.add_argument(
        "-g", "--grammar", const="_", nargs="?", help="Removed. Use -i/--inflect"
    )
    parser.add_argument(
        "-i", "--inflect", action="store_true", help=_("display inflections")
    )
    parser.add_argument(
        "--export",
        action="store_true",
        help=_("export parsed word attributes in yaml format"),
    )
    parser.add_argument(
        "--words-before", action="store_true", help=_("list 5 words before this one")
    )
    parser.add_argument(
        "--words-after", action="store_true", help=_("list 5 words after this one")
    )

    parser.add_argument(
        "-r",
        "--result",
        type=int,
        help=_(
            "display n-th (starting from 1) result in case "
            "of multiple words matching the input"
        ),
    )
    parser.add_argument(
        "--fuzzy", action="store_true", help=_("enable fuzzy word matching")
    )
    parser.add_argument(
        "--no-cache",
        action="store_false",
        dest="cache",
        help=_("do not cache retrieved words"),
    )

    parser.add_argument(
        "-V", "--version", action="store_true", help=_("print program version")
    )
    parser.add_argument(
        "--phonetic", action="store_true", help=_("display pronunciation")
    )
    parser.add_argument(
        "--alternative-spellings",
        action="store_true",
        help=_("display alternative spellings"),
    )
    args = parser.parse_args()

    if args.grammar:
        parser.error("The -g/--grammar was replaced with -i/--inflect .")
    return args


def main():
    """
    Take the first CLI argument and describe the corresponding word
    """

    # handle the --version switch
    if "--version" in sys.argv or "-V" in sys.argv:
        print("duden " + __version__)
        sys.exit(0)

    # parse normal arguments
    args = parse_args()

    # search all words matching the string
    words = search(
        args.word, return_words=False, exact=not args.fuzzy, cache=args.cache
    )

    # exit if the word wasn't found
    if not words:
        print(red(_("Word '{}' not found")).format(args.word))
        sys.exit(1)

    # list the options when there is more than one matching word
    if len(words) > 1 and args.result is None:
        print(
            _(
                "Found {} matching words. Use the -r/--result argument to "
                "specify which one to display."
            ).format(white(len(words), bold=True))
        )
        for i, word in enumerate(words, 1):
            print("{} {}".format(blue("{})".format(i)), word))
        sys.exit(1)

    result_index = args.result if args.result is not None else 1

    # choose the correct result
    try:
        word_url_suffix = words[result_index - 1]
    except IndexError:
        print(red(_("No result with number {}.")).format(result_index))
        sys.exit(1)

    # fetch and parse the word
    try:
        word = get(word_url_suffix, cache=args.cache)
    except Exception as exception:  # pylint: disable=broad-except
        print(red(exception))
        sys.exit(1)

    display_word(word, args)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
The duden package can parse the https://www.duden.de/ word information.

The `get` function is used to return parsed word, when provided with the word's
exact url name. The `search` function is used to search for words, either
returning exact matches (homonyms), or if fuzzy search is enabled, similar
words.

The basic class representing the parsed word is `DudenWord`.
"""
__all__ = [
    "get",
    "search",
    "get_word_of_the_day",
]

# grammatical categories enums
from .inflection import (
    Case,
    Degree,
    Gender,
    ImperativePerson,
    InfinitiveForm,
    Mood,
    Number,
    Person,
    Tense,
)
from .request import get, get_word_of_the_day, search



# -*- coding: utf-8 -*-
"""
Console printing functions
"""
from string import ascii_lowercase

import yaml
from crayons import blue, white, yellow  # pylint: disable=no-name-in-module


def display_inflections(word):
    """
    Display word's inflection table
    """
    inf_str = yaml.dump(
        word.inflection.data, indent=2, allow_unicode=True, sort_keys=False
    )
    print(inf_str)


def display_table(table, cell_spacing=" "):
    """
    Display general table with aligned columns
    """
    cols = list(zip(*table))
    collens = [max(len(word) for word in col) for col in cols]

    for row in table:
        for elem, collen in zip(row, collens):
            print(elem.ljust(collen), end=cell_spacing)
        print()


def display_compounds(word, compound_type="ALL"):
    """
    Print word common compounds
    """
    if compound_type == "ALL":
        for part_of_speech, compounds in word.compounds.items():
            print(white("# " + part_of_speech.capitalize(), bold=True))
            print_string_or_list(compounds)
            print()
    else:
        print_string_or_list(word.compounds[compound_type])


def print_tree_of_strings(tree):
    """
    Print a tree of strings up to depth 2

    Args:
        tree: tree of strings

    Example:

    >>> print_tree_of_strings(['Hase', ['Eins', 'Zwei']])
    0. Hase
    <BLANKLINE>
    1.  a. Eins
        b. Zwei
    """
    if isinstance(tree, str):
        print(tree)
        return

    for index, value in enumerate(tree):
        if isinstance(value, str):
            print(blue("{:>2}. ".format(index)), value, sep="")
        elif isinstance(value, list):
            for index_inner, value_inner in zip(ascii_lowercase, value):
                indent = (
                    blue("{:>2}. ".format(index)) if index_inner == "a" else " " * 4
                )
                print(
                    "{} {}".format(indent, blue(index_inner)),
                    blue(". "),
                    value_inner,
                    sep="",
                )
        print()


def print_string_or_list(obj):
    """
    Print string value or all list elements, if of type list
    """
    if isinstance(obj, list):
        for elem in obj:
            print(elem)
    else:
        print(obj)


def describe_word(word):
    """
    Print overall word description
    """
    # pylint: disable=too-many-branches
    print(yellow(word.title, bold=True))
    print(yellow("=" * len(word.title)))

    if word.phonetic:
        print(white(_("Pronunciation:"), bold=True), word.phonetic)
    if word.alternative_spellings:
        print(
            white(_("Alternative spellings:"), bold=True),
            ", ".join(word.alternative_spellings),
        )
    if word.part_of_speech:
        print(white(_("Word type:"), bold=True), word.part_of_speech)
    if word.usage:
        print(white(_("Usage:"), bold=True), word.usage)
    if word.frequency:
        commonness = "{label} {frequency}{max_frequency}".format(
            label=white(_("Commonness:"), bold=True),
            frequency=word.frequency,
            max_frequency=blue("/5"),
        )
        print(commonness)
    if word.word_separation:
        print(
            "{label} {content}".format(
                label=white(_("Separation:"), bold=True),
                content=str(blue("|")).join(word.word_separation),
            )
        )

    if word.origin:
        print(white(_("Origin:"), bold=True), word.origin)

    if word.grammar_overview:
        print(white(_("Grammar:"), bold=True), word.grammar_overview)

    if word.pronunciation_audio_url:
        print(
            "{label} {content}".format(
                label=white(_("Pronunciation audio:"), bold=True),
                content=white(word.pronunciation_audio_url, bold=False),
            )
        )

    if word.meaning_overview:
        print(white(_("Meaning overview:"), bold=True))
        print_tree_of_strings(word.meaning_overview)

    if word.synonyms:
        print(white(_("Synonyms:"), bold=True))
        print_tree_of_strings(word.synonyms)

    if word.compounds:
        print(white(_("Typical compounds:"), bold=True))
        for part_of_speech, words in word.compounds.items():
            print(blue(" - {}:".format(part_of_speech.capitalize())), ", ".join(words))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Contains the DudenWord class: a parser of duden.de response.
"""

import copy
import gettext
import os

from . import request  # pylint: disable=cyclic-import
from .common import clear_text, recursively_extract

EXPORT_ATTRIBUTES = [
    "name",
    "urlname",
    "title",
    "article",
    "part_of_speech",
    "usage",
    "frequency",
    "word_separation",
    "meaning_overview",
    "origin",
    "grammar_overview",
    "compounds",
    "synonyms",
    "words_before",
    "words_after",
    "phonetic",
    "alternative_spellings",
]

gettext.install("duden", os.path.join(os.path.dirname(__file__), "locale"))


class DudenWord:
    """
    Represents parsed word. Takes a BeautifulSoup object as a constructor argument.

    Example:

        > r = requests.get('https://www.duden.de/rechtschreibung/Hase')
        > soup = bs4.BeautifulSoup(r.text)
        > word = duden.DudenWord(soup)
        > word
        Hase, der (Substantiv, maskulin)
    """

    # pylint: disable=too-many-public-methods
    wordcloud_parts_of_speech = ["substantive", "verben", "adjektive"]

    def __init__(self, soup):
        self.soup = soup
        self._inflection = None

    def __repr__(self):
        return "{} ({})".format(self.title, self.part_of_speech)

    @property
    def title(self):
        """
        The word string with article
        """
        return self.soup.h1.get_text().replace("\xad", "").strip()

    @property
    def name(self):
        """
        Word without article
        """

        # Find span with class "lemma__main"
        title_element = self.soup.find("span", {"class": "lemma__main"})
        if title_element is not None:
            # remove soft hyphens "\xad" and return
            return clear_text(title_element.get_text())

        #  if the title_element does not exist, we fall back to the old method
        if self.part_of_speech is not None and "Substantiv" not in self.part_of_speech:
            return self.title
        if ", " not in self.title:
            return self.title

        name, _ = self.title.split(", ", 1)
        return name

    @property
    def urlname(self):
        """
        Return unique representation of the word used in duden.de urls
        """
        return self.soup.head.find("link", rel="canonical").attrs["href"].split("/")[-1]

    @property
    def revision_url(self):
        """Returns url to this specific word revision"""
        return self.soup.find("input", id="cite-field").attrs["value"]

    @property
    def node_no(self):
        """Returns word node number"""
        return self.revision_url.split("/")[-3]

    @property
    def revision_no(self):
        """Returns word revision number"""
        return self.revision_url.split("/")[-1]

    @property
    def article(self):
        """
        Word article
        """
        # Find span with class "lemma__determiner"
        article_element = self.soup.find("span", {"class": "lemma__determiner"})
        if article_element is not None:
            # remove soft hyphens "\xad" and return
            return clear_text(article_element.get_text())

        #  if the article_element does not exist, we fall back to the old method
        if self.part_of_speech is not None and "Substantiv" not in self.part_of_speech:
            return None
        if ", " not in self.title:
            return None

        _, article = self.title.split(", ", 1)
        return article

    def _find_tuple_dl(self, key, element=None):
        """
        Get value element corresponding to key element containing the text
        provided by the `key` argument
        """
        if element is None:
            element = self.soup.article

        dls = element.find_all("dl", class_="tuple", recursive=False)
        for dl_node in dls:
            label = dl_node.find("dt", class_="tuple__key")
            if key in label.text:
                return dl_node.find("dd", class_="tuple__val")

        return None

    @property
    def part_of_speech(self):
        """
        Return the part of speech
        """
        try:
            pos_element = self._find_tuple_dl("Wortart")
            return pos_element.text
        except AttributeError:
            return None

    @property
    def frequency(self):
        """
        Return word frequency:

        0 - least frequent
        5 - most frequent
        """
        try:
            freq_bar = self.soup.find("span", class_="shaft__full")
            return len(freq_bar.text)
        except AttributeError:
            return None

    @property
    def usage(self):
        """
        Return usage context
        """
        try:
            element = self._find_tuple_dl("Gebrauch")
            return element.text
        except AttributeError:
            return None

    @property
    def word_separation(self):
        """
        Return the word separated in a form of a list
        """
        containing_div = self.soup.find("div", id="rechtschreibung")
        sep_element = self._find_tuple_dl("Worttrennung", containing_div)
        if not sep_element:
            return None

        return sep_element.text.split("|")

    @property
    def pronunciation_audio_url(self):
        """
        Return the a url of a audio file for the word pronunciation.
        """
        audio_link = self.soup.select("a.pronunciation-guide__sound")
        if not audio_link:
            return None

        audio_link_href = str(audio_link[0].get("href"))

        return audio_link_href

    @property
    def meaning_overview(self):
        """
        Return the meaning structure, which can be string, list or a dict
        """
        section = self.soup.find("div", id="bedeutung") or self.soup.find(
            "div", id="bedeutungen"
        )
        if section is None:
            return None
        section = copy.copy(section)
        section.header.extract()

        # 1. remove examples
        for dl_node in section.find_all("dl", class_="note"):
            # pylint: disable=condition-evals-to-constant
            if True or dl_node.dt.text == "Beispiele":
                dl_node.extract()

        # 2. remove grammar parts
        for dl_node in section.find_all("dl", class_="tuple"):
            if dl_node.dt.text in ["Grammatik", "Gebrauch"]:
                dl_node.extract()

        # 3. remove pictures
        for node in section.find_all("figure"):
            node.extract()

        return recursively_extract(section, maxdepth=2, exfun=lambda x: x.text.strip())

    @property
    def synonyms(self):
        """
        Return the structure with word synonyms
        """
        section = self.soup.find("div", id="synonyme")
        if section is None:
            return None
        section = copy.copy(section)
        if section.header:
            section.header.extract()
        more_nav = section.find("nav", class_="more")
        if more_nav:
            more_nav.extract()

        return split_synonyms(section.text.strip())

    @property
    def origin(self):
        """
        Return the word origin
        """
        section = self.soup.find("div", id="herkunft")
        if section is None:
            return None

        section = copy.copy(section)
        if section.header:
            section.header.extract()
        return section.text.strip()

    @property
    def grammar_overview(self):
        """
        Return short grammar overview
        """
        section = self.soup.find("div", id="grammatik")
        if section is None:
            return None

        section = copy.copy(section)
        if section.header:
            section.header.extract()
        if section.nav:
            section.nav.extract()
        return section.text.strip() or None

    @property
    def compounds(self):
        """
        Return the typical word compounds
        """
        section = self.soup.find("div", id="kontext")
        if not section:
            return None

        pos_trans = {
            "noun": "substantive",
            "verb": "verben",
            "adj": "adjektive",
        }

        compounds = {}

        cluster_element = section.find("figure", class_="tag-cluster__cluster")
        for a_node in cluster_element.find_all("a"):
            compound_word = a_node.text
            compound_type = pos_trans[a_node.attrs["data-group"]]

            if compound_type not in compounds:
                compounds[compound_type] = []

            compounds[compound_type].append(compound_word)

        compounds_sorted = {}
        for pos in sorted(compounds.keys()):
            compounds_sorted[pos] = sorted(compounds[pos])

        return compounds_sorted

    @property
    def grammar(self):
        """Redirect users to new function"""
        raise RuntimeError("The .grammar property was replaced by .inflection")

    @property
    def grammar_raw(self):
        """Redirect users to new function"""
        raise RuntimeError("The .grammar_raw property was replaced by .inflection")

    @property
    def inflection(self):
        """
        Return word's Inflector object with methods for conjugation and declension

        This property performs a network request, so unless the request is cached, it
        takes a few seconds to return the result.
        """
        if self._inflection is None:
            self._inflection = (
                request.grammar(self.grammar_link) if self.grammar_link else None
            )

        return self._inflection

    @property
    def grammar_link(self):
        """
        Relative url for grammar table page, e.g. "/deklination/substantive/Petersilie"
        """
        section = self.soup.find("div", id="grammatik")
        if not section:
            return None

        link = section.find("a", id="grammatik")
        if not link:
            return None

        return link.attrs["href"]

    @property
    def can_decline(self):
        """Whether word provides declination data"""
        return self.grammar_link.startswith("/deklination")

    @property
    def can_conjugate(self):
        """Whether word provides conjugation data"""
        return self.grammar_link.startswith("/konjugation")

    def export(self):
        """
        Export word's attributes as a dictionary

        Used e.g. for creating test data.
        """
        worddict = {}
        for attribute in EXPORT_ATTRIBUTES:
            worddict[attribute] = getattr(self, attribute, None)
        worddict["inflection"] = self.inflection and self.inflection.data
        return worddict

    @property
    def before_after_structure(self):
        """
        Parsed "Blätter section"

        Returns: dict mapping section names to list of words tuples.
                 Each tuple is comprised of word name and word urlname.

        Example:
            >>> duden.get("laufen").before_after_structure
            {'Im Alphabet davor': [('Laufbekleidung', 'Laufbekleidung'),
              ('Laufbrett', 'Laufbrett'),
              ('Laufbursche', 'Laufbursche'),
              ('Läufchen', 'Laeufchen'),
              ('Läufel', 'Laeufel')],
             'Im Alphabet danach': [('laufend', 'laufend'),
              ('laufen lassen, laufenlassen', 'laufen_lassen'),
              ('Laufer', 'Laufer'),
              ('Läufer', 'Laeufer'),
              ('Lauferei', 'Lauferei')]}
        """
        result = {}
        section = self.soup.find("div", id="block-numero-beforeafterblock-2")
        for group in section.find_all("nav", class_="hookup__group"):
            h3title = group.h3.text
            result[h3title] = []
            for item in group.find_all("li"):
                link = item.a.attrs["href"].split("/")[-1]
                result[h3title].append((clear_text(item.text), link))
        return result

    @property
    def words_before(self):
        """Returns 5 words before this one in duden database"""
        return [name for name, _ in self.before_after_structure["Im Alphabet davor"]]

    @property
    def words_after(self):
        """Returns 5 words after this one in duden database"""
        return [name for name, _ in self.before_after_structure["Im Alphabet danach"]]

    @property
    def phonetic(self):
        """
        Returns pronunciation of the word in phonetic notation.
        See: https://en.wikipedia.org/wiki/International_Phonetic_Alphabet
        """
        ipa = self.soup.find("span", {"class": "ipa"})
        if ipa is not None:
            return ipa.get_text()

        return None

    @property
    def alternative_spellings(self):
        """
        Returns alternate spellings
        """
        alternative_spellings = self.soup.find_all(
            "span", {"class": "lemma__alt-spelling"}
        )
        if alternative_spellings is None:
            return None

        return [spelling.get_text() for spelling in alternative_spellings]


def split_synonyms(text):
    """
    Properly split strings like

    meaning1, (commonly) meaning2; (formal, distant) meaning3
    """
    # split by ',' and ';'
    comma_splits = text.split(",")
    fine_splits = []
    for split in comma_splits:
        fine_splits.extend(split.split(";"))

    # now join back parts which are inside of parentheses
    final_splits = []
    inside_parens = False
    for split in fine_splits:
        if inside_parens:
            final_splits[-1] = final_splits[-1] + "," + split
        else:
            final_splits.append(split)

        if "(" in split and ")" in split:
            inside_parens = split.index("(") > split.index("(")
        elif "(" in split:
            inside_parens = True
        elif ")" in split:
            inside_parens = False

    return [split.strip() for split in final_splits]
